# -*- coding: utf-8 -*-

"""Cloud API related exception"""

from requests import HTTPError

__all__ = ['LoginRequiredException']


class LoginRequiredException(HTTPError):
    """Basic username/password auth required."""



# -*- coding: utf-8 -*-

"""Cloud Token implementation"""

from __future__ import annotations

__all__ = ['NucleusToken']

import datetime
import os
import pickle  # nosec
import typing

import jwt

from anaconda_navigator.config.base import SUBFOLDER, get_home_dir
from .utilities import pickle as pickle_utils

if typing.TYPE_CHECKING:
    import requests


TOKEN_FILE: typing.Final[str] = os.path.join(os.path.join(get_home_dir(), SUBFOLDER), 'anaconda_nucleus.pkl')
TOKEN_LIFETIME: typing.Final[int] = 900
BASIC_TOKEN_TYPE: typing.Final[str] = 'Bearer'


EMPTY: typing.Any = object()


class NucleusToken:  # CloudToken
    """
    Represent tokens and metadata provided and required by Cloud API as a single object.

    :param username: Name of the logged user.
    :param access_token: Token to perform operations on Cloud API.
    :param refresh_token: Token used to refresh expired `access_token`.
    :param token_type: Type of `access_token`.
    :param expires_in: Value in seconds to calculate expiration date of `access_token` if `expiration_date` is not set.
    :param expiration_date: The expiration date of `access_token`.
    """

    __slots__ = ('__username', '__access_token', '__refresh_token', '__token_type', '__expiration_date')

    def __init__(  # pylint: disable=too-many-arguments
            self,
            username: typing.Optional[str] = None,
            access_token: typing.Optional[str] = None,
            refresh_token: typing.Optional[str] = None,
            token_type: str = BASIC_TOKEN_TYPE,
            expires_in: int = TOKEN_LIFETIME,
            expiration_date: typing.Optional[datetime.datetime] = None,
    ) -> None:
        """Initialize new :class:`~NucleusToken` instance."""
        if expiration_date is None:
            expiration_date = datetime.datetime.now() + datetime.timedelta(seconds=expires_in)

        self.__username: typing.Final[typing.Optional[str]] = username
        self.__access_token: typing.Final[typing.Optional[str]] = access_token
        self.__refresh_token: typing.Final[typing.Optional[str]] = refresh_token
        self.__token_type: typing.Final[str] = token_type
        self.__expiration_date: typing.Final[datetime.datetime] = expiration_date

    # def __getstate__(self) -> typing.Mapping[str, typing.Any]:
    #     """Serialize current instance."""
    #     return {
    #         'username': self.__username,
    #         'access_token': self.__access_token,
    #         'refresh_token': self.__refresh_token,
    #         'token_type': self.__token_type,
    #         'expiration_date': self.__expiration_date,
    #     }

    def __setstate__(
            self,
            state: typing.Union[
                None,
                typing.Mapping[str, typing.Any],
                typing.Sequence[typing.Optional[typing.Mapping[str, typing.Any]]],
            ],
    ) -> None:
        """Deserialize state of an instance."""
        if not isinstance(state, typing.Sequence):
            state = (state,)

        item: typing.Optional[typing.Mapping[str, typing.Any]]
        result: typing.Dict[str, typing.Any] = {}
        for item in state:
            if item is None:
                continue
            result.update({
                key.rsplit('__', 1)[-1]: value
                for key, value in item.items()
            })

        self.__init__(**result)  # type: ignore

    # Value properties

    @property
    def username(self) -> typing.Optional[str]:  # noqa: D401
        """Name of the logged user."""
        return self.__username

    @property
    def access_token(self) -> typing.Optional[str]:  # noqa: D401
        """Token to perform operations on Cloud API."""
        return self.__access_token

    @property
    def refresh_token(self) -> typing.Optional[str]:  # noqa: D401
        """Token used to refresh expired `access_token`."""
        return self.__refresh_token

    @property
    def token_type(self) -> str:  # noqa: D401
        """Type of `access_token`."""
        return self.__token_type

    @property
    def expiration_date(self) -> datetime.datetime:  # noqa: D401
        """The expiration date of `access_token`."""
        return self.__expiration_date

    # Calculated properties

    @property
    def expired(self) -> bool:  # noqa: D401
        """Token is past its expiration date."""
        return datetime.datetime.now() > self.expiration_date

    @property
    def valid(self) -> bool:  # noqa: D401
        """Token can be used for authentication."""
        return not (self.access_token is None or self.expired)

    @property
    def authorization(self) -> str:  # noqa: D401
        """Value for `AUTHORIZATION` header."""
        if self.valid:
            return f'{self.token_type} {self.access_token}'
        return ''

    # Modifications

    def replace(  # pylint: disable=too-many-arguments
            self,
            username: typing.Optional[str] = EMPTY,
            access_token: typing.Optional[str] = EMPTY,
            refresh_token: typing.Optional[str] = EMPTY,
            token_type: str = EMPTY,
            expires_in: int = EMPTY,
            expiration_date: typing.Optional[datetime.datetime] = EMPTY,
    ) -> NucleusToken:
        """
        Create a copy of current :class:`~NucleusToken` instance with optional modifications.

        :param username: Custom `username` value.
        :param access_token: Custom `access_token` value.
        :param refresh_token: Custom `refresh_token` value.
        :param token_type: Custom `token_type` value.
        :param expires_in: Custom `expires_in` value.
        :param expiration_date: Custom `expiration_date` value.
        :return: Copy of current instance with requested modifications.
        """
        if username is EMPTY:
            username = self.__username
        if access_token is EMPTY:
            access_token = self.__access_token
        if refresh_token is EMPTY:
            refresh_token = self.__refresh_token
        if token_type is EMPTY:
            token_type = self.__token_type
        if expires_in is EMPTY:
            expires_in = TOKEN_LIFETIME
            if expiration_date is EMPTY:
                expiration_date = self.__expiration_date
        elif expiration_date is EMPTY:
            expiration_date = None

        return NucleusToken(
            username=username,
            access_token=access_token,
            refresh_token=refresh_token,
            token_type=token_type,
            expires_in=expires_in,
            expiration_date=expiration_date,
        )

    def from_response(self, response: 'requests.Response', username: typing.Optional[str] = None) -> NucleusToken:
        """
        Import details from authentication HTTP response.

        :param response: Response to get details from.
        :param username: Optional username, which was used for login operation.
        :return: New token with imported details.
        """
        body: typing.Mapping[str, typing.Any] = response.json()
        changes: typing.Dict[str, typing.Any] = {
            'access_token': body['access_token'],
            'token_type': BASIC_TOKEN_TYPE,
        }

        if username is not None:
            changes['username'] = username

        refresh_token: typing.Optional[str] = response.cookies.get('refresh_token', None)
        if refresh_token:
            changes['refresh_token'] = refresh_token

        try:
            access_data: typing.Mapping[str, typing.Any] = jwt.decode(
                body['access_token'],
                options={'verify_signature': False},
            )
            changes['expiration_date'] = datetime.datetime.fromtimestamp(access_data['exp'])
        except (jwt.exceptions.PyJWTError, KeyError, TypeError):
            changes['expires_in'] = TOKEN_LIFETIME

        return self.replace(**changes)

    # Serialization

    @staticmethod
    def from_file(path: str = TOKEN_FILE) -> typing.Optional[NucleusToken]:
        """
        Update token attributes from the pickle file.

        :returns: :class:`~NucleusToken` instance, if requested file exists. :code:`None` otherwise.
        """
        try:
            stream: typing.BinaryIO
            with open(path, 'rb') as stream:
                result: typing.Any = pickle_utils.MigrationUnpickler(stream).load()
                if not isinstance(result, NucleusToken):
                    return None
                return result
        except (FileNotFoundError, pickle.PickleError):
            return None

    def save(self, path: str = TOKEN_FILE) -> None:
        """
        Dump token attributes to the pickle file.

        :param path: Path to file to save token to.
        """
        stream: typing.BinaryIO
        with open(path, 'wb') as stream:
            pickle.dump(self, stream)



# -*- coding: utf-8 -*-

"""Utility functions to use for application detection."""

from __future__ import annotations

__all__ = ['collect_str', 'join']

import os
import re
import typing
from .. import validation_utils
from . import folders

T_co = typing.TypeVar('T_co', covariant=True)
RecursiveOptionalString = typing.Union[None, str, 'RecursiveIterable[typing.Optional[str]]']


class RecursiveIterable(typing.Protocol[T_co]):  # pylint: disable=too-few-public-methods
    """Iterable that can yield another recursive iterable alongside other items."""

    def __iter__(self) -> typing.Iterator[typing.Union[T_co, RecursiveIterable[T_co]]]:
        """Iterate through a content of the object."""


def collect_str(source: RecursiveOptionalString) -> typing.Iterator[str]:
    """Collect string values from the recursive collections/iterables of :class:`~str` or :code:`None` instances."""
    stack: typing.List[RecursiveOptionalString] = [source]
    while stack:
        current: RecursiveOptionalString = stack.pop()

        if current is None:
            continue

        if isinstance(current, str):
            yield current

        else:
            offset: int = len(stack)
            child: typing.Union[None, str, typing.Iterable[typing.Any]]
            for child in current:
                stack.insert(offset, child)


@typing.overload
def join(root: str, *args: str) -> str:
    """Join path parts, if all of them are not :code:`None`."""


@typing.overload
def join(root: typing.Optional[str], *args: typing.Optional[str]) -> typing.Optional[str]:
    """Join path parts, if all of them are not :code:`None`."""


def join(root: typing.Optional[str], *args: typing.Optional[str]) -> typing.Optional[str]:
    """Join path parts, if all of them are not :code:`None`."""
    if root is None:
        return None

    if any(arg is None for arg in args):
        return None
    args = typing.cast(typing.Tuple[str, ...], args)

    return os.path.join(root, *args)


INLINE_VALUE_PATTERN: typing.Final[typing.Pattern[str]] = re.compile(r'{([^}]*)}')


def parse_and_join(value: typing.Union[str, typing.Sequence[str]]) -> typing.Optional[str]:
    """Parse path parts and join them."""
    if isinstance(value, str):
        return value

    # `str` is added here for the verbosity of the message
    validation_utils.of_type(str, typing.Sequence)(value)

    part: typing.Optional[str]
    parts: typing.List[typing.Optional[str]] = []
    for part in validation_utils.iterable_items(value):
        validation_utils.is_str(part)

        index: int
        items: typing.List[str] = INLINE_VALUE_PATTERN.split(part)
        for index in range(1, len(items), 2):
            item: 'folders.Folder' = typing.cast('folders.Folder', items[index].strip())
            validation_utils.of_options(*folders.FOLDERS.keys())(item, field_name=('inline', index // 2))
            current: typing.Optional[str] = folders.FOLDERS[item]
            if current is None:
                parts.append(None)
                break
            items[index] = current
        else:
            parts.append(''.join(items))

    validation_utils.has_items(at_least=1)(parts)

    with validation_utils.catch_exception():
        return join(*parts)  # pylint: disable=no-value-for-parameter



# -*- coding: utf-8 -*-

"""Solvers for invalid default_env setting."""

from __future__ import annotations

__all__ = ()

import html
import os
import typing
from . import core


@core.CONFIGURATION_POOL.register
def solve_missing_environment(context: core.ConfigurationContext) -> typing.Optional[core.SolvedError]:
    """Check if `default_env` is missing from file system."""
    default_env: str = context.config.get('main', 'default_env')
    if default_env and os.path.isdir(default_env):
        return None

    context.config.set('main', 'default_env', context.api.ROOT_PREFIX)

    if default_env:
        return core.SolvedError(
            caption='Broken Navigator configuration',
            message=f'Environment is no longer available:<br>- {html.escape(default_env)}',
            tags='default_env',
        )

    return None


@core.CONFLICT_POOL.register
def solve_unknown_environment(context: core.ConflictContext) -> typing.Optional[core.SolvedError]:
    """Check if `default_env` is unknown to conda."""
    default_env: str = context.config.get('main', 'default_env')
    if default_env in context.conda_info['processed_info']['__environments']:
        return None

    context.config.set('main', 'default_env', context.api.ROOT_PREFIX)

    if default_env:
        return core.SolvedError(
            caption='Broken Navigator configuration',
            message=f'Environment is no longer available:<br>- {html.escape(default_env)}',
            tags='default_env',
        )

    # Just in case, as it should not trigger under normal circumstances
    return core.SolvedError(
        caption='Broken Navigator configuration',
        message='Default environment is updated',
        tags='default_env',
    )



# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Community Tab."""

from __future__ import annotations

__all__ = ['CommunityTab']

import collections
import contextlib
import hashlib
import json
import os
import re
import typing

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

from anaconda_navigator.config import CONF, CONTENT_PATH, IMAGE_DATA_PATH
from anaconda_navigator.static.content import LINKS_INFO_PATH
from anaconda_navigator.static.images import LOGO_PATH, VIDEO_ICON_PATH
from anaconda_navigator.utils import download_manager
from anaconda_navigator.utils import url_utils
from anaconda_navigator.utils.logs import logger
from anaconda_navigator.widgets import ButtonBase, FrameTabContent, FrameTabHeader, SpacerHorizontal, WidgetBase
from anaconda_navigator.widgets.helperwidgets import LineEditSearch
from anaconda_navigator.widgets.lists.content import ListItemContent, ListWidgetContent


def parse_content(download: download_manager.Download) -> typing.Tuple[bool, typing.List[typing.Any]]:
    """
    Parse data fetched from https://anaconda.com/api/content.

    Returns tuple of primary content flag and the content itself. Primary content flag is used as a sign that primary
    content is loaded and doesn't need to be fetched from the file.
    """
    content: typing.List[typing.Any] = []
    try:
        stream: typing.TextIO
        with download.open('rt', encoding='utf-8') as stream:
            content = json.load(stream)
    except (OSError, TypeError, ValueError):
        pass
    return bool(content), content


def parse_events(download: download_manager.Download) -> typing.Tuple[bool, typing.List[typing.Any]]:
    """
    Parse data fetched from https://anaconda.com/api/events.

    Output is similar to :func:`~parse_content`.
    """
    content: typing.List[typing.Any] = []
    try:
        stream: typing.TextIO
        with download.open('rt', encoding='utf-8') as stream:
            content = json.load(stream)
    except (OSError, TypeError, ValueError):
        pass

    item: typing.Any
    for item in content:
        item['tags'] = ['event']
        item['uri'] = item.pop('url', '')

    return False, content[:25]


def parse_videos(download: download_manager.Download) -> typing.Tuple[bool, typing.List[typing.Any]]:
    """
    Parse data fetched from https://anaconda.com/api/videos.

    Output is similar to :func:`~parse_content`.
    """
    content: typing.List[typing.Any] = []
    try:
        stream: typing.TextIO
        with download.open('rt', encoding='utf-8') as stream:
            content = json.load(stream)
    except (OSError, TypeError, ValueError):
        pass

    item: typing.Any
    for item in content:
        item['tags'] = ['video']
        item['uri'] = item.pop('video', '')
        if item['uri']:
            item['banner'] = item.pop('thumbnail', '')
        item['date'] = item.pop('date_start', '')

    return False, content[:25]


def parse_webinars(download: download_manager.Download) -> typing.Tuple[bool, typing.List[typing.Any]]:
    """
    Parse data fetched from https://anaconda.com/api/webinars.

    Output is similar to :func:`~parse_content`.
    """
    content: typing.List[typing.Any] = []
    try:
        stream: typing.TextIO
        with download.open('rt', encoding='utf-8') as stream:
            content = json.load(stream)
    except (OSError, TypeError, ValueError):
        pass

    item: typing.Any
    for item in content:
        item['tags'] = ['webinar']
        item['uri'] = item.pop('url', '')

        image: typing.Any = item.pop('image', '')
        if image and isinstance(image, dict):
            item['banner'] = image.get('src', '')

    return False, content[:25]


def combine(group: download_manager.Group) -> typing.Tuple[bool, typing.Any]:
    """
    Combine outputs from all data sources into a single content

    Output is similar to :func:`~parse_content`.
    """
    total_flag: bool = False
    total_content: typing.Any = []

    download: download_manager.Download
    for download in group.downloads:
        if download.result is None:
            continue

        current_flag: bool
        current_content: typing.Any
        try:
            current_flag, current_content = download.result
        except (TypeError, ValueError):
            continue

        total_flag = total_flag or current_flag
        total_content.extend(current_content)

    return total_flag, total_content


# --- Widgets used in CSS styling
# -----------------------------------------------------------------------------

class ButtonToggle(ButtonBase):
    """Toggle button used in CSS styling."""

    def __init__(self, *args, **kwargs):
        """Toggle button used in CSS styling."""
        super().__init__(*args, **kwargs)
        self.setCheckable(True)
        self.clicked.connect(lambda v=None: self._fix_check)

    def _fix_check(self):
        self.setProperty('checked', self.isChecked())
        self.setProperty('unchecked', not self.isChecked())


# --- Main widgets
# -----------------------------------------------------------------------------

def image_file_from_banner(banner: str) -> str:
    """Retrieve `image_file` value from a `banner`."""
    return f'{hashlib.md5(banner.encode("utf-8")).hexdigest()[-8:]}_{url_utils.file_name(banner)}'  # nosec


class CommunityTab(WidgetBase):  # pylint: disable=too-many-instance-attributes
    """Tab with a community content."""

    sig_ready = QtCore.Signal()

    def __init__(  # pylint: disable=too-many-arguments
            self,
            parent: typing.Optional[QtCore.QObject] = None,
            tags: typing.Optional[typing.Sequence[str]] = None,
            utm_medium: str = '',
    ) -> None:
        """Initialize new :class:`~CommunityTab` content."""
        super().__init__(parent=parent)

        self.__content: typing.List[typing.Any] = []
        self.__content_offset: int = 0
        self.__content_step: int = 1
        self.__tags: typing.Optional[typing.Sequence[str]] = tags

        self.__pixmaps: typing.Dict[str, QtGui.QPixmap] = {}
        self.__filter_widgets: typing.List[ButtonToggle] = []
        self.__default_pixmap = QtGui.QPixmap(VIDEO_ICON_PATH).scaled(
            100,
            60,
            QtCore.Qt.KeepAspectRatio,
            QtCore.Qt.FastTransformation,
        )

        # Widgets
        self.setObjectName('Tab')

        self.__timer_load: QtCore.QTimer = QtCore.QTimer()
        self.__timer_load.setInterval(333)

        self.__filters_layout = QtWidgets.QHBoxLayout()

        self.__text_filter: LineEditSearch = LineEditSearch()
        self.__text_filter.setPlaceholderText('Search')
        self.__text_filter.setAttribute(QtCore.Qt.WA_MacShowFocusRect, False)
        font_metrics: QtGui.QFontMetrics = self.__text_filter.fontMetrics()
        self.__text_filter.setMaximumWidth(font_metrics.width('M' * 23))

        layout_header: QtWidgets.QHBoxLayout = QtWidgets.QHBoxLayout()
        layout_header.addLayout(self.__filters_layout)
        layout_header.addStretch()
        layout_header.addWidget(self.__text_filter)

        frame_header: FrameTabHeader = FrameTabHeader()
        frame_header.setLayout(layout_header)

        self.__list: ListWidgetContent = ListWidgetContent(utm_medium=utm_medium)
        self.__list.setAttribute(QtCore.Qt.WA_MacShowFocusRect, False)
        self.__list.setMinimumHeight(200)

        layout_content: QtWidgets.QHBoxLayout = QtWidgets.QHBoxLayout()
        layout_content.addWidget(self.__list)

        frame_content: FrameTabContent = FrameTabContent()
        frame_content.setLayout(layout_content)

        layout: QtWidgets.QVBoxLayout = QtWidgets.QVBoxLayout()
        layout.addWidget(frame_header)
        layout.addWidget(frame_content)
        self.setLayout(layout)

        # Signals
        self.__timer_load.timeout.connect(self.__put_content)
        self.__text_filter.textChanged.connect(self.__filter_content)

    def setup(self) -> None:
        """Set up tab content."""
        group = download_manager.Group(
            download_manager.Download('https://anaconda.com/api/events').into(
                os.path.join(CONTENT_PATH, 'events.json'),
            ).process(
                parse_events,
            ),
            download_manager.Download('https://anaconda.com/api/videos').into(
                os.path.join(CONTENT_PATH, 'videos.json'),
            ).process(
                parse_videos,
            ),
            download_manager.Download('https://anaconda.com/api/webinars').into(
                os.path.join(CONTENT_PATH, 'webinars.json'),
            ).process(
                parse_webinars,
            ),
            download_manager.Download('https://anaconda.com/api/content').into(
                os.path.join(CONTENT_PATH, 'content.json'),
            ).process(
                parse_content,
            ),
        ).require(
            download_manager.Require.NONE,
        ).process(
            combine,
        )

        group.sig_succeeded.connect(lambda _: self.__process_content(group=group))

        download_manager.manager().execute(group)

    def __process_content(self, group: download_manager.Group) -> None:
        """Apply content to the tab after it is loaded from the web."""
        flag: bool
        content: typing.List[typing.Any]
        flag, content = group.result

        stream: typing.TextIO
        if not flag:
            try:
                with open(LINKS_INFO_PATH, 'rt', encoding='utf-8') as stream:
                    content.extend(json.load(stream))
            except (OSError, ValueError):
                logger.exception('failed to load bundled community content')

        item: typing.Any
        for item in content:
            item['uri'] = item['uri'].replace('<p>', '').replace('</p>', '').replace(' ', '%20')

            banner: str = item.get('banner', '')
            if banner:
                item['image_file'] = image_file_from_banner(banner)

            item['image_file_path'] = os.path.join(IMAGE_DATA_PATH, item.get('image_file', '...'))

        content.sort(key=lambda tile: (tile.get('sticky', '') not in (True, 'True', 'true', '1'), tile.get('tags', [])))
        self.__content = content

        # with open(CONTENT_JSON_PATH, 'wt', encoding='utf-8') as stream:
        #     json.dump(content, stream)

        self.__make_tag_filters()
        self.__timer_load.start(31)

    def __make_tag_filters(self) -> None:
        """Create tag filtering checkboxes based on available content tags."""
        counts: typing.Counter[str] = collections.Counter(
            tag
            for item in self.__content
            for tag in item.get('tags', [])
        )

        if self.__tags is None:
            self.__tags = sorted(counts)

        for tag in sorted(self.__tags):
            count: int = counts[tag]
            if not count:
                continue

            toggle = ButtonToggle(f'{tag.capitalize()} ({count})'.strip())
            toggle.setObjectName(tag.lower())
            toggle.setChecked(CONF.get('checkboxes', tag.lower(), True))
            toggle.clicked.connect(self.__filter_content)

            self.__filter_widgets.append(toggle)
            self.__filters_layout.addWidget(toggle)
            self.__filters_layout.addWidget(SpacerHorizontal())

    def __put_content(self) -> None:
        """
        Add items to the list.

        This method adds only a part of the content. Each next call adds next part to the interface.
        """
        index: int
        for index in range(self.__content_offset, self.__content_offset + self.__content_step):
            if index >= len(self.__content):
                self.__timer_load.stop()
                self.sig_ready.emit()
                break

            item: typing.Any = self.__content[index]

            banner: str = item.get('banner', None) or ''
            path: str = item.get('image_file_path', None) or ''
            content_item: ListItemContent = ListItemContent(
                title=item.get('title', None) or '',
                subtitle=item.get('subtitle', None) or '',
                uri=item.get('uri', None) or '',
                date=item.get('date', None) or '',
                summary=item.get('summary', None) or '',
                tags=item.get('tags', []),
                banner=banner,
                path=path,
                pixmap=self.__default_pixmap,
            )
            self.__list.addItem(content_item)

            # This allows the content to look for the pixmap
            content_item.pixmaps = self.__pixmaps

            # Use images shipped with Navigator, if no image try the download
            local_image: str = os.path.join(LOGO_PATH, item.get('image_file', '...'))
            if os.path.isfile(local_image):
                self.__process_thumbnail(content_item, local_image)
            elif banner and path:
                self.__download_thumbnail(content_item, banner, path)

        self.__content_offset += self.__content_step
        self.__filter_content()

    def __download_thumbnail(self, item: ListItemContent, url: str, path: str) -> None:
        """Fetch thumbnail for a card from the web."""
        download: download_manager.Download
        download = download_manager.Download(url).into(path).keep(download_manager.Keep.FRESH)
        download.sig_succeeded.connect(lambda: self.__process_thumbnail(item, path))
        download_manager.manager().execute(download)

    def __process_thumbnail(self, item: ListItemContent, path: str) -> None:
        """Process thumbnail and apply it to a card."""
        if path not in self.__pixmaps:
            try:
                if not os.path.isfile(path):
                    self.__pixmaps[path] = QtGui.QPixmap()
                else:
                    extension: str = os.path.splitext(path)[1][1:].upper()
                    if extension in ('PNG', 'JPG', 'JPEG'):
                        self.__pixmaps[path] = QtGui.QPixmap(path, format=extension)
                    else:
                        self.__pixmaps[path] = QtGui.QPixmap(path)
            except OSError:
                logger.exception('failed to initialize community card thumbnail')

        with contextlib.suppress(RuntimeError):
            item.update_thumbnail(self.__pixmaps[path])

    def __filter_content(self):
        """Filter content by a search string on all the fields of the item."""
        tokens: typing.List[str] = list(
            filter(bool, map(str.strip, re.split(r'\W', self.__text_filter.text().lower()))),
        )

        toggle: ButtonToggle
        selected_tags: typing.List[str] = []
        for toggle in self.__filter_widgets:
            tag: str = toggle.objectName()
            checked: bool = toggle.isChecked()

            CONF.set('checkboxes', tag, checked)
            if checked:
                selected_tags.append(tag)

        index: int
        for index in range(self.__list.count()):
            item = self.__list.item(index)

            all_checks: typing.List[bool] = []

            token: str
            for token in tokens:
                all_checks.append(
                    token in item.title.lower() or
                    token in item.venue.lower() or
                    token in ' '.join(item.authors).lower() or
                    token in item.summary.lower()
                )

            all_checks.append(any(
                tag.lower() in selected_tags
                for tag in item.tags
            ))

            item.setHidden(not all(all_checks))

    def ordered_widgets(
            self,
            next_widget: typing.Optional[QtWidgets.QWidget] = None,  # pylint: disable=unused-argument
    ) -> typing.List[QtWidgets.QWidget]:
        """Fix tab order of UI widgets."""
        return [
            *self.__filter_widgets,
            self.__text_filter,
            *self.__list.ordered_widgets(),
        ]

    def update_style_sheet(self) -> None:
        """Update custom CSS stylesheet."""
        self.__list.update_style_sheet()
